/*****************************************************************
Autor: 			Thomas B.
Beschreibung: 	DCF-Empfnger
Version:		1.4.3

Bekannte Bugs: keine

Verbesserungsmglichkeit: 	
*****************************************************************/
#include "../Libs/dcf77.h"

// eine 8 Bit Variable
struct 
{
	volatile unsigned char p_min	: 1	;		// Parity minute
	volatile unsigned char p_datum	: 1	;		// Parity Datum
	volatile unsigned char p_stunde	: 1	;		// Parity stunde
	volatile unsigned char synchronisiert:1;	// zeigt an ob Uhr schon einmal DCF-synchronisiert wurde
												// --> wird fr Funktion dcf_vergleich bentigt
	
	volatile unsigned int dcf_error	:1; // 1 Bit fr dcf_error
	volatile unsigned int sync		:3;	// 3 Bit fr sync
										//	flags.sync = 0 // nix synchronisiert
										//	flags.sync = 1 // sync-Beginn / minutenanfang
										//	flags.sync = 2 // minute komplett empfangen / korrekt aufgezeichnet
										//	flags.sync = 3 // Datum & Uhrzeit korrekt berechnet --> kann bernommen
										//	werden! --> nchster Schritt nach bernahme wre wieder sync = 1
} flags;

// 8-Bit Variable
volatile uint8_t 	sek 			= 0,	//	Sekunde 	(Timergesteuert)
					min 			= 0, 	//	Minute 		(Timergesteuert)
					stunde 			= 0,	//	Stunde 		(Timergesteuert)
					tag 			= 0,	//	Tag 		(Timergesteuert)
					wochentag 		= 0,	//	Wochentag 	(Timergesteuert)
					monat 			= 0,	//	Monat 		(Timergesteuert)	
					jahr 			= 0,	//	Jahr		(Timergesteuert)
					dcf_sek 		= 0,	//	Sekunde 	(DCF-gesteuert)
					dcf_min 		= 0,	//	Minute 		(DCF-gesteuert)
					dcf_stunde 		= 0,	//	Stunde 		(DCF-gesteuert)
					dcf_tag 		= 0,	//	Tag 		(DCF-gesteuert)
					dcf_wochentag 	= 0,	//	Wochentag 	(DCF-gesteuert)
					dcf_monat 		= 0,	//	Monat 		(DCF-gesteuert)
					dcf_jahr 		= 0,	//	Jahr		(DCF-gesteuert)
					verstr_min		= 0,	//	Zhler wieviele Minuten seit der letzten 
											//	Synchronisation verstrichen sind
					sync_timer 		= 0;	//	Zhlt Zeit in Sekunden. Alle 60 sek: verstr_min++
					
// 16-Bit Variable				
volatile uint16_t	dcf_timer 		= 0,	//	Zhlt wieviele Timerberlufe zwischen den Flanken
											//	stattgefunden haben --> dient zur DCF-Auswertung
					watch_timer 	= 0;	//	Timer fr Offline-Zeitberechnung. Zhlt Anzahl der Timerberlufe
											//	Wenn Anzahl der Timerberlufe == Sek dann wird Uhrzeit weitergezhlt

/****************************************************************
Aktiviert den externen Interrupt --> DCF-Synchronisation aktiv
****************************************************************/
void int_ext0_enable(void)
{
	GICR |= (1<<INT0); // Ext INT auf 1;
}

/****************************************************************
berprft ob der externe Interrupt aktiviert ist
****************************************************************/
int is_int_enabled(void)
{
	if(((GICR >> INT0)&0x1) == 1)
		return 1;
	else
		return 0;
}

/****************************************************************
Deaktiviert den externen Interrupt --> DCF-Synchronisation inaktiv
****************************************************************/
void int_ext0_disable(void)
{
	GICR &= ~(1<<INT0); // Ext INT auf 0;
}

/****************************************************************
Initialisiert alle Variablen, die zur Synchronisation notwendig sind
****************************************************************/
void dcf_erase(void)
{
	dcf_sek				= 0;
	dcf_min 			= 0;
	dcf_stunde 			= 0;
	dcf_tag 			= 0;
	dcf_wochentag 		= 0;
	dcf_monat 			= 0;
	dcf_jahr 			= 0;
	dcf_timer 			= 0;
	flags.dcf_error 	= 0;
	flags.sync 			= 0;
	int_to_display(flags.sync,1,1,10);
	flags.p_datum		= 0;
	flags.p_min			= 0;
	flags.p_stunde		= 0;
}

/****************************************************************
Grundgedanke: 	Wird eine Uhr nicht regelmig, bzw. in groen Zeitabstnden
				synchronisiert, kann es aufgrund der Ungenauigkeit des Quarzes
				bzw. sonstigen Strungen zu Ungenauigkeiten kommen --> Sek luft
				vor oder nach.
				Hier setzt die Funktion an. Wenn die Synchronisation lnger her 
				ist wird nach erfolgreicher Neusynchronisation die empfangene
				Zeit mit der intern hochgezhlten verglichen. Weichen diese voneinander
				ab, wird ein Korrekturwert fr das Compare Register berechnet und
				im Compare Register abgelegt.
****************************************************************/
void dcf_vergleich(void)
{
	uint32_t korrekturwert = 0;
	uint16_t reg = 0/*, akt_timerwert = 0*/;
	
	reg = ((unsigned int) OCR1AH<<8) | (unsigned int) OCR1AL;			//Compare Register - Wie hoch zhlt der Timer
//	akt_timerwert = ((unsigned int) TCNT1H<<8) | (unsigned int) TCNT1L;	//Wo steht der Timer gerade?

	if(min >= dcf_min)
	{
		korrekturwert = (((min-dcf_min)*60*SEKUNDE + sek*SEKUNDE + watch_timer)*reg/* + akt_timerwert*/) / (verstr_min*60*SEKUNDE);
		reg= reg+korrekturwert;
	}
	else
	{
		korrekturwert = ((dcf_min - min)*60*SEKUNDE*reg - sek*SEKUNDE*reg - watch_timer*reg /*+ akt_timerwert*/) / (verstr_min*60*SEKUNDE);
		reg= reg-korrekturwert;
	}

	OCR1AH = (reg >> 8) & 0xFF;
	OCR1AL = (reg & 0xFF);
	int_to_display(reg, 5,4,15);
	int_to_display(korrekturwert, 3,3,15);
}

/****************************************************************
Beschreibung:	Funktion zhlt die berechnete Uhrzeit hoch - diese  
				Funktion ist DCF-Empfangs-unabhngig.
****************************************************************/
void uhr_anzeige(void)
{
	uint8_t schaltjahr = 0;
	
	sek++;	
	
	if (sek == 60)
	{
		min++;
		
		if(min == 60)
		{
			stunde++;
			
			if(stunde == 24)
			{
				tag++;
				wochentag++;

				if (wochentag == 8)
					wochentag = 1;
			
				if (((monat == 1) || (monat == 3) || (monat == 5) || (monat == 7) || (monat == 8) || (monat == 10) || (monat == 12)) && (tag == 32))
				{
					if ( monat == 12)
					{
						monat = 1;
						jahr++;
						
						int_to_display(jahr,2,4,6);
					}
					else
						monat++;
					
					tag = 1;
				}
				else if (((monat == 4) || (monat == 6 || (monat == 9)) || (monat == 11)) && (tag == 31))
				{
						tag = 1;
						monat++;
				}
				else if (monat == 2)
				{
					if (((jahr%4 == 0) && (jahr%100 != 0)) || (jahr%400 ==0))
						schaltjahr = 1;		//Schaltjahr ist durch 4, aber nicht durch 100 teilbar. Nur Vielfache von 400 sind doch Schaltjahre!
		
					if (((schaltjahr == 1) && (tag == 30)) || ((schaltjahr != 1) && (tag == 29)))
					{
						tag = 1;
						monat++;
					}
				}

				int_to_display(wochentag,2,4,10);
				int_to_display(monat,2,4,3);
				int_to_display(tag,2,4,0);
				stunde = 0;	
			}

			int_to_display(stunde,2,3,0);
			min = 0;
		}			
		
		int_to_display(min,2,3,3);
		sek = 0;
	}

	int_to_display(sek,2,3,6);
}


/****************************************************************
Beschreibung: 	Funktion bernimmt die vom DCF empfangenen und
				ausgewerteten Uhrzeit & Datumsinformationen
****************************************************************/
void dcf_uebernahme(void)
{
	watch_timer = 0;
	
	if(sek != dcf_sek)
	{		
		sek = dcf_sek;
		int_to_display(sek,2,3,6);
	}
	
	if(min != dcf_min)
	{
		min = dcf_min;
		int_to_display(min,2,3,3);
	}
	
	if(stunde != dcf_stunde)
	{
		stunde = dcf_stunde;
		int_to_display(stunde,2,3,0);
	}
	
	if(tag != dcf_tag)
	{
		tag = dcf_tag;
		int_to_display(tag,2,4,0);
	}
	
	if(wochentag != dcf_wochentag)
	{
		wochentag = dcf_wochentag;
		int_to_display(wochentag,2,4,10);
	}
	
	if(monat != dcf_monat)
	{
		monat = dcf_monat;
		lcd_putc_pos(0x2E,4,2);
		int_to_display(monat,2,4,3);
	}
	
	if(jahr != dcf_jahr)
	{
		jahr = dcf_jahr;
		lcd_putc_pos(0x2E,4,5);
		int_to_display(jahr,2,4,6);
	}
	
	dcf_erase();
}

/****************************************************************
Timerberlauf
****************************************************************/
ISR(SIG_OUTPUT_COMPARE1A)
{	
	cli();
	dcf_timer++;	
	watch_timer++;
	
	// Anzahl der berluft pro Sekunde --> SEKUNDE = (F_CPU/10000)
	if(watch_timer == SEKUNDE)
	{
		watch_timer = 0;		
		uhr_anzeige();

		if((DCF_SYNC_TIMER != 0) && (flags.synchronisiert ==1))
			sync_timer++;
	}

	// Nach 3 Sekunden: Signal verloren
	if(dcf_timer == 3*SEKUNDE)
		dcf_erase();

	if((sync_timer == 60) && (DCF_SYNC_TIMER != 0) && (flags.synchronisiert ==1))
	{
		verstr_min++;
		
		int_to_display(DCF_SYNC_TIMER,3,1,15);
		int_to_display(verstr_min,3,2,15);

		if ((verstr_min+1) == DCF_SYNC_TIMER)
			int_ext0_enable();
		
		if(verstr_min == 255)
			verstr_min--;
		
		sync_timer = 0;
	}

	sei();
}

/****************************************************************
Externer Interrupt mit DCF-Algorithmus
****************************************************************/
ISR(SIG_INTERRUPT0)
{
	cli();
	
	//Timerzeit berechnet sich durch (F_CPU/10000) * 5% --> (F_CPU/10000)*5/100 --> ZEHNTEL = (F_CPU/10000)/100
	if ((flags.sync == 1) && (dcf_timer >= (ZEHNTEL*5)) && (dcf_timer <= (ZEHNTEL*15)))		// == logische "0"
	{
		dcf_timer = 0;
		
		if (dcf_sek == 58)
		{
			flags.sync = 2;
			int_to_display(flags.sync,1,1,10);
		}
	}
	else if ((flags.sync == 1) && (dcf_timer > (ZEHNTEL*15)) && (dcf_timer <= (ZEHNTEL*25)))	// == logische "1"
	{
		dcf_timer = 0;

		if ((dcf_sek >= 21 ) && ( dcf_sek <= 27))		// Minute
			dcf_min |= (1 << (dcf_sek - 21));
		else if (dcf_sek == 28)							// Paritt Minute
			flags.p_min = 1;
		else if ((dcf_sek >= 29 ) && ( dcf_sek <= 34))	// Stunde
			dcf_stunde |= (1 << (dcf_sek - 29));
		else if (dcf_sek == 35)							// Paritt Stunde
			flags.p_stunde = 1;
		else if ((dcf_sek >= 36 ) && ( dcf_sek <= 41))	// Tag
			dcf_tag |= (1 << (dcf_sek - 36));
		else if ((dcf_sek >= 42 ) && ( dcf_sek <= 44))	// Wochentag
			dcf_wochentag |= (1 << (dcf_sek - 42));
		else if ((dcf_sek >= 45 ) && ( dcf_sek <= 49))	// Monat
			dcf_monat |= (1 << (dcf_sek - 45));
		else if ((dcf_sek >= 50 ) && ( dcf_sek <= 57))	// Jahr
			dcf_jahr |= (1 << (dcf_sek - 50));
		else if (dcf_sek == 58 ) 						// Paritt Datum (Tag & Wochentag & Monat & Jahr)
			flags.p_datum = 1;
		
		if (dcf_sek == 58)
		{
			flags.sync = 2;
			int_to_display(flags.sync,1,1,10);
		}
	}
	else if ((flags.sync == 1) && (dcf_timer >= (ZEHNTEL*75)) && (dcf_timer <= (ZEHNTEL*95)))	// == nix
	{
		dcf_timer = 0;
		dcf_sek ++;

		int_to_display(dcf_sek,2,2,10);
	}
	else if (dcf_timer >= (ZEHNTEL*120))	// == Sekunde 58 & 59 --> Synchronistationssekunde
	{			
		dcf_timer = 0;
		dcf_sek = 0;	// Minutenanfang, d.h. sek = 0;	
		
		// bernahme der empfangenen & berechneten DCF-Daten
		if (flags.sync == 3)	
		{
			sync_timer = 0;
			
			// Falls die empfangenen Sekunden und Minuten-Informationen mit den berechneten
			// nicht bereinstimmen wird Funktion dcf_vergleich aufgerufen um die Abweichung 
			// berechnen und das Compare-Register anzupassen.
			if (flags.synchronisiert == 1)
				dcf_vergleich();

			flags.synchronisiert = 1;	// zeigt an ob schon einmal synchronisiert wurde
			
			dcf_uebernahme();	// bernahme der Empfangenen DCF-Daten
			
			// Falls der DCF-Empfang nur in Zeitintervallen erfolgen soll, wird hier
			// der erforderliche externe Interrupt deaktiviert.
			if (DCF_SYNC_TIMER != 0)	
				int_ext0_disable();
			
			verstr_min = 0;
			int_to_display(verstr_min,3,2,15);		
		}

		flags.sync = 1;
		int_to_display(flags.sync,1,1,10);
	}
	else
	{
		dcf_timer = 0;
		flags.dcf_error = 1;
	}
	
	// Auswertung nach sek 58!
	if (flags.sync == 2)
	{
		unsigned int 	parity_min = 0,	parity_h = 0,parity_date = 0;
					
		// Parittsberechnung der Minutenbits --> XOR Verknfpung: 1^1 = 0, 1^0 = 1, 0^0 = 0;	
		for(int i = 0; i<=6; i++)
			parity_min ^= ((dcf_min>>i)&0x1);
		
		//Parittsbit ausfiltern == ((dcf_stunde>>4)&0x1)
		if(parity_min == flags.p_min)
	/*		Bits werden in Hex-Form bertragen --> Beispiel uhrzeit.minute 22:
			Binr: 0010 0010 (==20 + 2) == Dez: 34 == Hex 22
			Berechnung: 1.	(0010 0010)>> 4				= 0010 = 2
						2. 	0010 * 10 					= 20
						3.	20 + (0010 0010 & 0000 1111) = 20 + 2 = 22
	*/		
			dcf_min = ((dcf_min>>4)*10)+(dcf_min&0xF);
		else
			flags.dcf_error = 1;
		
		// Parittsberechnung der Stundenbits --> XOR Verknfpung: 1^1 = 0, 1^0 = 1, 0^0 = 0;
		for(int i = 0; i<=5; i++)
			parity_h ^= ((dcf_stunde>>i)&0x1);
		
		if(parity_h == flags.p_stunde)
			dcf_stunde = ((dcf_stunde>>4)*10)+(dcf_stunde&0xF);	
		else
			flags.dcf_error = 1;

		// Parittsberechnung Datum --> XOR
		for(int i = 0; i<=5 ; i++)
			parity_date ^= ((dcf_tag>>i)&0x1);
		for(int i = 0; i<=2 ; i++)
			parity_date ^= ((dcf_wochentag>>i)&0x1);
		for(int i = 0; i<=4 ; i++)
			parity_date ^= ((dcf_monat>>i)&0x1);
		for(int i = 0; i<=7 ; i++)
			parity_date ^= ((dcf_jahr>>i)&0x1);

		if(parity_date == flags.p_datum)
		{
			//tag - 6 Bit
			dcf_tag = ((dcf_tag>>4)*10)+(dcf_tag&0xF);
			//wochentag - 3 Bit
			dcf_wochentag = ((dcf_wochentag>>4)*10)+(dcf_wochentag&0xF);
			//Kalendermonat - 5 Bit
			dcf_monat = ((dcf_monat>>4)*10)+(dcf_monat&0xF);
			//jahr - 8 Bit
			dcf_jahr = ((dcf_jahr>>4)*10)+(dcf_jahr&0xF);
		}
		else
			flags.dcf_error = 1;
			
		flags.sync = 3;			//Berechnung abgeschlossen --> bernahme der Daten mglich
		int_to_display(flags.sync,1,1,10);
	}
	
	if (flags.dcf_error == 1)
		dcf_erase();
		
	sei();
}


/****************************************************************
****************************************************************/
void timer1_init(void)
{
	// 16 - Bit Timer
	sei();

	//Timer/Counter1 Control Register A
	TCCR1A = 0x0;
	//Timer/Counter1 Control Register B
	TCCR1B = 0x9;		// CTC Mode, Prescaler 1

	//Output Compare Register 1 A
	// Timer zhlt bis 9999 (0x270F) -> 0 - 9999 == 10000
	OCR1AH = 0x27;
	OCR1AL = 0x0F;
	
	TIMSK = 0x10;
}


/****************************************************************
Funktion wandelt einen int-wert in char um und gibt ihn an der geforderten
Position aus (y = Zeile, x = Spalte).
"laenge" gibt eine Mindestlnge fr das ausgegebene Zeichen an
Bsp.: int_to_display(9,2,1,0)
--> Ausgabe: 09
Bsp int_to_display(3001,2,1,0)
-->.: Ausgabe: 3001

n		= die in char umzuwandelnde int-Gre
laenge	= Mindestlnge fr das ausgegebene Char-Zeichen
y		= Zeilenangabe fr Ausgabe auf Display
x		= Spaltenangabe fr Ausgabe auf Display
****************************************************************/
void int_to_display(int n, int laenge,int y, int x)
{
	int i = 0, j = 0, k = 0;
	char temp = 0, *s;
	
	s = &temp;	
	
	// Prfen auf negatives int
	if (n < 0) 
	{
		n = -n;
        i++;
        k = 1;
	}
    
    //ermittelt Anzahl der Zeichen von n
	for (j = n;j > 0 ;i++, j/=10);	
	
	//Zeichenanzahl von n kleiner als die vorgegebene Laenge?
	if (laenge > i)
	{
		s[laenge] = '\0';
		
		if (i == 0)
			i = 1;
        
		for (j = 0;j<=(laenge-i);s[j++]='0');

		// int ist negativ
        if (k == 1)
			s[laenge-i] = '-';		
		
		i = laenge-1;       
	}
	else
	{
		s[i--] = '\0';
		
		// int ist negativ
		if(k == 1)
			s[0] = '-';
	}
	
    // int in Zeichenkette kopieren
	for(j = i;n>0;n/=10)
		s[j--] = n % 10 + '0';

	lcd_puts_pos(s,y,x);
/*	
Alternativen:
	// 2 Zeichen ausgeben
	// lcd_putc_pos((char)((i/10 + 0x30)&0xFF),3,6);
	// lcd_putc_pos((char)(((i-(i/10)*10)+0x30)&0xFF),3,7);
	
	// 3 Zeichen ausgeben
	//	lcd_putc_pos((i/100 + 0x30),2,0);
	//	lcd_putc_pos(((((i-(i/100)*100)/10)) + 0x30),2,1);		
	//	lcd_putc_pos(((i-(i/10)*10)+0x30),2,2);
*/
}


/****************************************************************
****************************************************************/
void dcf_init(void)
{
	lcd_puts_pos("00:00:00",3,0);
	
	timer1_init();
	MCUCR = 0x1; // Externer Interrupt @ any logical change
	int_ext0_enable();	
	
	dcf_erase();

	flags.synchronisiert = 0;	
}

